class Condition:
    def __init__(self, typ, name, value):
        self.typ = typ
        self.name = name
        self.value = value

    def __str__(self):
        if isinstance(self.value, str):
            return f"{self.name}:\"{self.value}\""
        return f"{self.name}:{self.value}"


class AndCondition(Condition):
    def __init__(self, name, value):
        super().__init__(',', name, value)


class Query:

    def __init__(self, query_obj=None):
        self.__queries = []
        if query_obj:
            for k, v in query_obj.items():
                self.query(k, v)

    def query(self, name, value):
        self.__queries.append(AndCondition(name, value))
        return self

    def generate(self):
        ret = ''
        if len(self.__queries) > 0:
            ret += str(self.__queries[0])
        if len(self.__queries) > 1:
            for e in self.__queries[1:]:
                ret += f"{e.typ}{e}"
        return ret

    def __str__(self):
        return self.generate()



class AuthException(Exception):
    pass


class ResponseException(Exception):
    pass


class ParentNotFoundException(Exception):
    pass


class EmptyException(Exception):
    pass


class ResourceNotFound(Exception):
    pass


class NotAnInstanceException(Exception):
    pass



class DefaultVersionTemplate:

    def generate(self, api, project_version_id):
        self.api = api
        self.project_version_id = project_version_id

        return [
            self.attributes(),
            self.responsibilities(),
            self.copy_from_partial(),
            self.copy_project_version_state(),
            self.configure_processing_rules(),
            self.commit_project_version()
        ]

    def attributes(self):
        return self.api.construct_request('PUT', f"/api/v1/projectVersions/{self.project_version_id}/attributes", [
            {
                "attributeDefinitionId": 5,
                "values": [
                    {
                        "guid": "New"
                    }
                ],
                "value": None
            },
            {
                "attributeDefinitionId": 6,
                "values": [
                    {
                        "guid": "Internal"
                    }
                ],
                "value": None
            },
            {
                "attributeDefinitionId": 7,
                "values": [
                    {
                        "guid": "internalnetwork"
                    }
                ],
                "value": None
            },
            {
                "attributeDefinitionId": 1,
                "values": [
                    {
                        "guid": "High"
                    }
                ],
                "value": None
            },
        ])

    def responsibilities(self):
        return self.api.construct_request('PUT', f"/api/v1/projectVersions/{self.project_version_id}/responsibilities", [
            {
                "responsibilityGuid": "projectmanager",
                "userId": None
            },
            {
                "responsibilityGuid": "securitychampion",
                "userId": None
            },
            {
                "responsibilityGuid": "developmentmanager",
                "userId": None
            }
        ])

    def copy_from_partial(self):
        return self.api.construct_request('POST', f"/api/v1/projectVersions/{self.project_version_id}/action", [
            {
                "type": "COPY_FROM_PARTIAL",
                "values": {
                    "projectVersionId": self.project_version_id,
                    "previousProjectVersionId": -1,
                    "copyAnalysisProcessingRules": True,
                    "copyBugTrackerConfiguration": True,
                    "copyCurrentStateFpr": False,
                    "copyCustomTags": True
                }
            }
        ])

    def commit_project_version(self):
        return self.api.construct_request('PUT', f"/api/v1/projectVersions/{self.project_version_id}", {
            'committed': True
        })

    def copy_project_version_state(self):
        return self.api.construct_request('POST', f"/api/v1/projectVersions/{self.project_version_id}/action", [
            {
                "type": "COPY_CURRENT_STATE",
                "values": {
                    "projectVersionId": self.project_version_id,
                    "previousProjectVersionId": -1,
                    "copyCurrentStateFpr": False
                }
            }
        ])

    def configure_processing_rules(self):
        return self.api.construct_request('PUT', f"/api/v1/projectVersions/{self.project_version_id}/resultProcessingRules", [
            {
                "identifier": "com.fortify.manager.BLL.processingrules.BuildProjectProcessingRule",
                "displayName": "Require approval if the Build Project is different between scans",
                "displayable": True,
                "enabled": False
            },
            {
                "identifier": "com.fortify.manager.BLL.processingrules.ExternalListVersionProcessingRule",
                "displayName": "Check external metadata file versions in scan against versions on server.",
                "displayable": True,
                "enabled": True
            },
            {
                "identifier": "com.fortify.manager.BLL.processingrules.FileCountProcessingRule",
                "displayName": "Require approval if file count differs by more than 10%",
                "displayable": True,
                "enabled": False
            },
            {
                "identifier": "com.fortify.manager.BLL.processingrules.ForceMigrationProcessingRule",
                "displayName": "Perform Force Instance ID migration on upload",
                "displayable": True,
                "enabled": False
            },
            {
                "identifier": "com.fortify.manager.BLL.processingrules.FortifyAnnotationsProcessingRule",
                "displayName": "Require approval if result has Fortify Java Annotations",
                "displayable": True,
                "enabled": False
            },
            {
                "identifier": "com.fortify.manager.BLL.processingrules.LOCCountProcessingRule",
                "displayName": "Require approval if line count differs by more than 10%",
                "displayable": True,
                "enabled": False
            },
            {
                "identifier": "com.fortify.manager.BLL.processingrules.MigrationProcessingRule",
                "displayName": "Automatically perform Instance ID migration on upload",
                "displayable": True,
                "enabled": True
            },
            {
                "identifier": "com.fortify.manager.BLL.processingrules.NewerEngineVersionProcessingRule",
                "displayName": "Require approval if the engine version of a scan is newer than the engine version of the previous scan",
                "displayable": True,
                "enabled": False
            },
            {
                "identifier": "com.fortify.manager.BLL.processingrules.QuickScanProcessingRule",
                "displayName": "Ignore SCA Scans performed in QuickScan mode",
                "displayable": True,
                "enabled": True
            },
            {
                "identifier": "com.fortify.manager.BLL.processingrules.RulePackVersionProcessingRule",
                "displayName": "Require approval if the rulepacks used in the scan do not match the rulepacks used in the previous scan",
                "displayable": True,
                "enabled": False
            },
            {
                "identifier": "com.fortify.manager.BLL.processingrules.ValidCertificationProcessingRule",
                "displayName": "Require approval if SCA or WebInspect Agent scan does not have valid certification",
                "displayable": True,
                "enabled": False
            },
            {
                "identifier": "com.fortify.manager.BLL.processingrules.WarningProcessingRule",
                "displayName": "Require approval if result has analysis warnings",
                "displayable": True,
                "enabled": False
            },
            {
                "identifier": "com.fortify.manager.BLL.processingrules.UnknownOrDisallowedAuditedAttrChecker",
                "displayName": "Warn if audit information includes unknown custom tag",
                "displayable": True,
                "enabled": True
            },
            {
                "identifier": "com.fortify.manager.BLL.processingrules.AuditedAnalysisRule",
                "displayName": "Require the issue audit permission to upload audited analysis files",
                "displayable": True,
                "enabled": True
            },
            {
                "identifier": "com.fortify.manager.BLL.processingrules.PendingApprovalChecker",
                "displayName": "Disallow upload of analysis results if there is one pending approval",
                "displayable": True,
                "enabled": False
            },
            {
                "identifier": "com.fortify.manager.BLL.processingrules.VetoCascadingApprovalProcessingRule",
                "displayName": "Disallow approval for processing if an earlier artifact requires approval",
                "displayable": True,
                "enabled": False
            }
        ])


class CloneVersionTemplate(DefaultVersionTemplate):

    def __init__(self, previous_version_id):
        self.previous_version_id = previous_version_id

    def copy_from_partial(self):
        return self.api.construct_request('POST', f"/api/v1/projectVersions/{self.project_version_id}/action", [
            {
                "type": "COPY_FROM_PARTIAL",
                "values": {
                    "projectVersionId": self.project_version_id,
                    "previousProjectVersionId": self.previous_version_id,
                    "copyAnalysisProcessingRules": True,
                    "copyBugTrackerConfiguration": True,
                    #"copyCurrentStateFpr": True, # didn't see this in 21.1?
                    "copyCustomTags": True,
                    "copyUserAccessSettings": True,
                    "copyVersionAttributes": True
                }
            }
        ])

    def copy_project_version_state(self):
        return self.api.construct_request('POST', f"/api/v1/projectVersions/{self.project_version_id}/action", [
            {
                "type": "COPY_CURRENT_STATE",
                "values": {
                    "projectVersionId": self.project_version_id,
                    "previousProjectVersionId": self.previous_version_id
                }
            }
        ])



import requests
from typing import Union, Tuple, Any
from .exceptions import *
from . import __version__


class FortifySSCAPI:
    """
    API object to talk to SSC via REST
    """

    def __init__(self, url: str,  auth: Union[str, Tuple[str, str]], proxies=None, verify=True):
        """
        :param url: url to ssc, including the path. E.g. `https://fortifyssc/ssc`
        :param auth: Authentication, either a token str or a (username, password) tuple
        """
        self.url = url.rstrip('/')
        self._token = None
        self.__user = None
        self.__pass = None
        self.proxies = proxies
        self.verify = verify

        if isinstance(auth, str):
            self._token = auth
        elif len(auth) == 2:
            # (user,pass)
            self.__user = auth[0]
            self.__pass = auth[1]
        else:
            raise AuthException("No valid authentication information")

    def __enter__(self):
        if self._token is None:
            self._authorize()
        return self

    def _authorize(self):
        self._token = self.create_token()

    def __exit__(self, type, value, traceback):
        if self.__user and self.__pass:
            self._unauthorize()

    def _unauthorize(self):
        self.revoke_token(self._token)
        self._token = None

    def create_token(self, description: str = 'PySSC Python API', type: str = 'UnifiedLoginToken') -> str:
        """
        Create a token with given username and password. Does not need to be manually called for API to work.
        Must use username:password authentication for this to work

        :param description: The description the token is created with
        :param type: The type of token to create, default being `UnifiedLoginToken`
        :raises AuthException: If we are unable to create a token
        :returns: The token itself
        """
        assert self.__user and self.__pass, "Cannot use token based authentication to create tokens."
        kwargs = dict(
            auth=(self.__user, self.__pass),
            json=dict(type=type, description=description)
        )
        if self.proxies:
            kwargs['proxies'] = self.proxies
        r = requests.post(f"{self.url}/api/v1/tokens", **kwargs)
        if r.status_code != 201:
            raise AuthException(f"Failed to authenticate - {r.status_code} - {r.text}")
        # could store user from the response, not sure why though
        return r.json()['data']['token']

    def revoke_token(self, token: str) -> None:
        """
        Revoke the given token
        Must use username:password authentication for this to work

        :param token: The token to revoke
        :raises AuthException: If we are unable to revoke the token
        """
        assert self.__user and self.__pass, "Cannot use token based authentication to create tokens."
        kwargs = dict(
            auth=(self.__user, self.__pass),
            json=dict(tokens=[token])
        )
        if self.proxies:
            kwargs['proxies'] = self.proxies
        r = requests.post(f"{self.url}/api/v1/tokens/action/revoke", **kwargs)
        if r.status_code != 200:
            raise AuthException(f"Failed to revoke token - {r.status_code} - {r.text}")

    def construct_request(self, method: str, path: str, data: Any) -> dict:
        """
        Construct a request for the bulk request API

        :param method: The HTTP method
        :param path: The relative path to the API endpoint (the `uri` argument)
        :param data: the postData
        """
        return {
            'uri': f"{self.url}/{path.lstrip('/')}",
            'httpVerb': method,
            'postData': data
        }

    def bulk_request(self, reqs):
        """
        :param reqs: The requests generated by py:func:`construct_request`
        """
        return self.post('/api/v1/bulk', requests=reqs)

    def page_data(self, endpoint, **kwargs):
        if 'start' not in kwargs:
            kwargs['start'] = 0
        if 'limit' not in kwargs:
            kwargs['limit'] = 200  # default

        r = self.get(endpoint, **kwargs)

        for e in r['data']:
            yield e

        data_len = len(r['data'])
        count = r['count']

        if (data_len + kwargs['start']) < count:
            kwargs['start'] = kwargs['start'] + kwargs['limit']
            for e in self.page_data(endpoint, **kwargs):
                yield e

    def get(self, endpoint, *args, **kwargs):
        """
        The available query parameters are:

        * q - field name + value pair that filters results
        * fields - comma-delimited list of fields to include in response
        * embed - item detail links to traverse and embed in the current response
        * orderby - field name by which to order results
        * start - used by paging to specify first item to retrieve
        * limit - used by paging to determine page size (defaults to 200 items)
        * groupby - used to group results

        :param endpoint:
        :param args: Args as a dict
        :param kwargs: The query parameters
        :return:
        """
        data = args[0] if args else {}
        data = {**data, **kwargs}
        return self._request('get', endpoint, params=data)

    def post(self, endpoint, *args, **kwargs):
        data = args[0] if args else {}
        data = {**data, **kwargs}
        return self._request('post', endpoint, json=data)

    def put(self, endpoint, *args, **kwargs):
        data = args[0] if args else {}
        data = {**data, **kwargs}
        print(data)
        return self._request('put', endpoint, json=data)

    def delete(self, endpoint, *args, **kwargs):
        data = args[0] if args else {}
        data = {**data, **kwargs}
        return self._request('delete', endpoint, params=data)

    def _request(self, method: str, endpoint: str, **kwargs):
        headers = {
            "Authorization": f"FortifyToken {self._token}",
            "Accept": 'application/json',
            "User-Agent": f"fortifyapi {__version__}"
        }
        if self.proxies:
            kwargs['proxies'] = self.proxies
        if not self.verify:
            kwargs['verify'] = self.verify
        r = requests.request(method, f"{self.url}/{endpoint.lstrip('/')}", headers=headers, **kwargs)
        if 200 <= r.status_code >= 299:
            if r.status_code == 409:
                raise ResourceNotFound(f"ResponseException - {r.status_code} - {r.text}")
            raise ResponseException(f"ResponseException - {r.status_code} - {r.text}")
        data = r.json()
        #print(f"{method} {endpoint}\n\t{r.text}")
        return data



#!/usr/bin/env python

import os
import sys

from fortifyapi import __version__ as version

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

with open('README.rst', 'r') as f:
    readme = f.read()

# Publish helper
if sys.argv[-1] == 'build':
    os.system('python setup.py sdist bdist')
    sys.exit(0)

if sys.argv[-1] == 'tag':
    os.system("git tag -a %s -m 'version %s'" % (version, version))
    os.system("git push --tags")
    sys.exit(0)

if sys.argv[-1] == 'publish-test':
    os.system('python setup.py sdist bdist_wheel upload -r pypitest')
    sys.exit(0)

setup(
    name='fortifyapi',
    packages=['fortifyapi'],
    version=version,
    description='Python library for Fortify Software Security Center (SSC) RESTFul API',
    long_description=readme,
    long_description_content_type = "text/x-rst",
    author='Brandon Spruth, Matthew Gill',
    author_email='brandon@spruth.co, mgill@c0ffee.me',
    url='https://github.com/fortifyadmin/fortifyapi',
    download_url='https://github.com/fortifyadmin/fortifyapi/tarball/' + version,
    license='MIT',
    zip_safe=True,
    test_suite='nose.collector',
    tests_require=['nose'],
    install_requires=['requests'],
    keywords=['fortify', 'api', 'security', 'software', 'microfocus', 'ssc', 'sast'],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'License :: OSI Approved :: MIT License',
        'Topic :: Software Development',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Software Development :: Libraries :: Application Frameworks',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ]
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

__author__ = "Brandon Spruth (brandon@spruth.co)"
__contributors__ = ["Brandon Spruth"]
__status__ = "Production"
__license__ = "MIT"

import urllib3
import json
import ntpath
import requests
import requests.auth
import requests.exceptions
import urllib.parse
from . import __version__ as version


class FortifyApi(object):
    def __init__(self, host, username=None, password=None, token=None, verify_ssl=True, timeout=60, user_agent=None,
                 client_version='21.10'):

        self.host = host
        self.username = username
        self.password = password
        self.token = token
        self.verify_ssl = verify_ssl
        self.timeout = timeout
        self.client_version = client_version

        if not user_agent:
            self.user_agent = 'fortify_api/' + version
        else:
            self.user_agent = user_agent

        if not self.verify_ssl:
            urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)

        # Set auth_type based on what's been provided
        if username is not None:
            self.auth_type = 'basic'
        elif token is not None:
            self.auth_type = 'token'
        else:
            self.auth_type = 'unauthenticated'

    def bulk_create_new_application_version_request(self, version_id, development_phase, development_strategy,
                                                    accessibility, business_risk_ranking, custom_attributes=[]):
        """
        Creates a new Application Version by using the Bulk Request API. 'create_new_project_version' must be used
        before calling this method.
        :param version_id: Version ID
        :param development_phase: Development Phase GUID of Version
        :param development_strategy: Development Strategy GUID of Version
        :param accessibility: Accessibility GUID of Version
        :param business_risk_ranking: Business Risk Rank GUID of Version
        :param custom_attributse: List of custom Attribute tuples that consist of attributeDefinitionId, values, & value. 
                                  Default is a empty string tuple.
        :return: A response object containing the newly created project and project version
        """
        data = self._bulk_format_new_application_version_payload(version_id=version_id,
                                                                 development_phase=development_phase,
                                                                 development_strategy=development_strategy,
                                                                 accessibility=accessibility,
                                                                 business_risk_ranking=business_risk_ranking,
                                                                 custom_attributes=custom_attributes)
        url = '/api/v1/bulk'
        return self._request('POST', url, data=data)

    @staticmethod
    def _bulk_format_attribute_definition(attribute_definition_id_value, guid_value='', value='null'):
        json_application_version = dict(attributeDefinitionId=attribute_definition_id_value,
                                        values=[],
                                        value=value)
        if guid_value is not None:
            json_application_version['values'] = [dict(guid=guid_value)]
        return json_application_version

    def _bulk_format_new_application_version_payload(self, version_id, development_phase, development_strategy,
                                                     accessibility, business_risk_ranking, custom_attributes):
        json_application_version = dict(requests=[
            self._bulk_create_attributes(version_id, development_phase, development_strategy, accessibility,
                                         business_risk_ranking, custom_attributes),
            self._bulk_create_responsibilities(version_id),
            self._bulk_create_configurations(version_id),
            self._bulk_create_commit(version_id),
            self._bulk_create_version(version_id)
        ])
        return json.dumps(json_application_version)

    def _bulk_create_attributes(self, version_id, development_phase, development_strategy,
                                accessibility, business_risk_ranking, custom_attributes):
        if business_risk_ranking is None:
            business_risk_ranking = 'High'
        json_application_version = dict(
            uri=self.host + '/api/v1/projectVersions/' + str(version_id) + '/attributes',
            httpVerb='PUT',
            postData=[
                self._bulk_format_attribute_definition('5', development_phase),
                self._bulk_format_attribute_definition('6', development_strategy),
                self._bulk_format_attribute_definition('7', accessibility),
                self._bulk_format_attribute_definition('1', business_risk_ranking),
            ])
        for a in custom_attributes:
            guid_value = a[1]
            if guid_value == "":
                guid_value = None
            json_application_version['postData'].append(
                self._bulk_format_attribute_definition(
                    attribute_definition_id_value=a[0],
                    guid_value=guid_value,
                    value=a[2]))

        return json_application_version

    def _bulk_create_responsibilities(self, version_id):
        json_application_version = dict(
            uri=self.host + '/api/v1/projectVersions/' + str(version_id) + '/responsibilities',
            httpVerb='PUT',
            postData=[]
        )
        json_application_version['postData'] = [dict(responsibilityGuid='projectmanager',
                                                     userId='null'),
                                                dict(responsibilityGuid='securitychampion',
                                                     userId='null'),
                                                dict(responsibilityGuid='developmentmanager',
                                                     userId='null'),
                                                ]
        return json_application_version

    def _bulk_create_configurations(self, version_id):
        json_application_version = dict(uri=self.host + '/api/v1/projectVersions/' + str(version_id) + '/action',
                                        httpVerb='POST',
                                        postData=[dict(
                                            type='COPY_FROM_PARTIAL',
                                            values={
                                                "projectVersionId": str(version_id),
                                                "previousProjectVersionId": '-1',
                                                "copyAnalysisProcessingRules": 'true',
                                                "copyBugTrackerConfiguration": 'true',
                                                "copyCurrentStateFpr": 'false',
                                                "copyCustomTags": 'true'
                                            }
                                        )]
                                        )
        return json_application_version

    def _bulk_create_commit(self, version_id):
        json_application_version = dict(
            uri=self.host + '/api/v1/projectVersions/' + str(version_id),
            httpVerb='PUT',
            postData={
                "committed": 'true'
            }
        )
        return json_application_version

    def _bulk_create_version(self, version_id):
        json_application_version = dict(uri=self.host + '/api/v1/projectVersions/' + str(version_id) + '/action',
                                        httpVerb='POST',
                                        postData=[dict(
                                            type='COPY_CURRENT_STATE',
                                            values={
                                                "projectVersionId": str(version_id),
                                                "previousProjectVersionId": '-1',
                                                "copyCurrentStateFpr": 'false'
                                            }
                                        )]
                                        )
        return json_application_version

    def set_processing_rules(self, version_id):
        """
        :param version_id: SSC Project Version to modify
        :return: A response object changing all required processing default fields from true to false
        """
        data = [
            {"identifier": "com.fortify.manager.BLL.processingrules.ExternalListVersionProcessingRule",
            "enabled": False},
            {"identifier": "com.fortify.manager.BLL.processingrules.FortifyAnnotationsProcessingRule",
             "enabled": False},
            {"identifier": "com.fortify.manager.BLL.processingrules.LOCCountProcessingRule",
             "enabled": False},
            {"identifier": "com.fortify.manager.BLL.processingrules.NewerEngineVersionProcessingRule",
             "enabled": False},
            {"identifier": "com.fortify.manager.BLL.processingrules.RulePackVersionProcessingRule",
             "enabled": False},
            {"identifier": "com.fortify.manager.BLL.processingrules.ValidCertificationProcessingRule",
             "enabled": False},
            {"identifier": "com.fortify.manager.BLL.processingrules.WarningProcessingRule",
             "enabled": False},
            {"identifier": "com.fortify.manager.BLL.processingrules.AuditedAnalysisRule",
             "enabled": False},
            {"identifier": "com.fortify.manager.BLL.processingrules.PendingApprovalChecker",
             "enabled": False},
            {"identifier": "com.fortify.manager.BLL.processingrules.QuickScanProcessingRule",
             "enabled": False},
            {"identifier": "com.fortify.manager.BLL.processingrules.FileCountProcessingRule",
             "enabled": False},
            {"identifier": "com.fortify.manager.BLL.processingrules.BuildProjectProcessingRule",
             "enabled": False},
            {"identifier": "com.fortify.manager.BLL.processingrules.VetoCascadingApprovalProcessingRule",
              "enabled": False},
            {"identifier": "com.fortify.manager.BLL.processingrules.PendingApprovalChecker",
              "enabled": False},
            {"identifier": "com.fortify.manager.BLL.processingrules.UnknownOrDisallowedAuditedAttrChecker",
              "enabled": False},
            { "identifier": "com.fortify.manager.BLL.processingrules.MigrationProcessingRule",
              "enabled": False} 
        ]
        url = '/api/v1/projectVersions/' + str(version_id) + '/resultProcessingRules'
        return self._request('PUT', url, json=data)
    
    def get_processing_rules(self, version_id):
        """
        :param version_id: Project Version ID from SSC to query
        :return: Listing of all result processing rules for a given SSC project version
        """
        url = '/api/v1/projectVersions/' + str(version_id) + '/resultProcessingRules'
        return self._request('GET', url)

    def create_application_version(self, application_name, application_template, version_name, description,
                                   application_id=None, issue_template_id=None):
        """
        :param application_name: Project name
        :param application_id: Project ID
        :param application_template: Application template name
        :param version_name: Version name
        :param description: Application Version description
        :param issue_template_id: Optional for when you already know the id you want to use
        :return: A response object containing the created project version
        """
        # If no application ID is given, sets JSON value to null.
        if application_id is None:
            application_id = 'null'

        if issue_template_id is None:
            # Gets Template ID
            issue_template = self.get_issue_template_id(project_template_name=application_template)
            issue_template_id = issue_template.data['data'][0]['id']

        data = dict(name=version_name,
                    description=description,
                    active=True,
                    committed=False,
                    project={
                        'name': application_name,
                        'description': description,
                        'issueTemplateId': issue_template_id,
                        'id': application_id
                    },
                    issueTemplateId=issue_template_id)

        url = '/api/v1/projectVersions'
        return self._request('POST', url, json=data)
    
    def delete_application_version(self, id):
        """
        Delete a given application or project version from SSC
        :param id: Project Version ID
        :return:
        """
        url = "/api/v1/projectVersions/" + str(id)
        return self._request('DELETE', url)
    
    def download_artifact(self, artifact_id):
        """
        You might use this method like this, for example
            api = FortifyApi("https://my-fortify-server:my-port", token=get_token())
            response, file_name = api.download_artifact_scan("my-id")
            if response.success:
                file_content = response.data
                with open('/path/to/some/folder/' + file_name, 'wb') as f:
                    f.write(file_content)
            else:
                print response.message

        We've coded this for the entire file to load into memory. A future change may be to permit
        streaming/chunking of the file and handing back a stream instead of content.
        :param artifact_id: the id of the artifact to download
        :return: binary file data and file name
        """
        file_token = self.get_file_token('DOWNLOAD').data['data']['token']

        url = "/download/artifactDownload.html?mat=" + file_token + "&id=" + str(
            artifact_id) + "&clientVersion=" + self.client_version

        headers = {
            'Accept': 'text/html,application/xhtml+xml,application/xml;q=0.9,image/webp,image/apng,*/*;q=0.8',
            'Accept-Encoding': 'gzip, deflate, br',
            'Connection': 'keep-alive'
        }

        response = self._request('GET', url, stream=True, headers=headers)

        try:
            file_name = response.headers['Content-Disposition'].split('=')[1].strip("\"'")
        except:
            file_name = ''

        return response, file_name

    def download_artifact_scan(self, artifact_id):
        """
        You might use this method like this, for example
            api = FortifyApi("https://my-fortify-server:my-port", token=get_token())
            response, file_name = api.download_artifact_scan("my-id")
            if response.success:
                file_content = response.data
                with open('/path/to/some/folder/' + file_name, 'wb') as f:
                    f.write(file_content)
            else:
                print response.message

        We've coded this for the entire file to load into memory. A future change may be to permit
        streaming/chunking of the file and handing back a stream instead of content.
        :param artifact_id: the id of the artifact scan to download
        :return: binary file data and file name
        """
        file_token = self.get_file_token('DOWNLOAD').data['data']['token']

        url = "/download/currentStateFprDownload.html?mat=" + file_token + "&id=" + str(
            artifact_id) + "&clientVersion=" + self.client_version + "&includeSource=true"

        headers = {
            'Accept': 'text/html,application/xhtml+xml,application/xml;q=0.9,image/webp,image/apng,*/*;q=0.8',
            'Accept-Encoding': 'gzip, deflate, br',
            'Connection': 'keep-alive'
        }

        response = self._request('GET', url, stream=True, headers=headers)

        try:
            file_name = response.headers['Content-Disposition'].split('=')[1].strip("\"'")
        except:
            file_name = ''

        return response, file_name

    def get_artifact_scans(self, parent_id):
        """
        :param parent_id: parent resource identifier
        :return: A response object containing artifact scans
        """
        url = "/api/v1/artifacts/" + str(parent_id) + "/scans"
        return self._request('GET', url)

    def get_attribute_definition(self, search_expression):
        """
        :param search_expression: A fortify-formatted search expression, e.g. Development Phase
        :return: A response object containing the result of the GET
        """
        if search_expression:
            url = '/api/v1/attributeDefinitions?q=name:"' + search_expression + '"'
            return self._request('GET', url)
        else:
            return FortifyResponse(message='A search expression must be provided', success=False)

    def get_attribute_definitions(self):
        """
        :return: A response object containing all attribute definitions
        """
        url = '/api/v1/attributeDefinitions?start=0&limit=-1'
        return self._request('GET', url)

    def get_cloudscan_jobs(self):
        """
        :return: A response object containing all cloudscan jobs
        """
        url = '/api/v1/cloudjobs?start=0&limit=-1'
        return self._request('GET', url)

    def get_cloudscan_job_status(self, scan_id):
        """
        :return: A response object containing a cloudscan job
        """
        url = '/api/v1/cloudjobs/' + scan_id
        return self._request('GET', url)

    def get_file_token(self, purpose):
        """
        :param purpose: specify if the token is for file 'UPLOAD' or 'DOWNLOAD'
        :return: a response body containing a file token for the specified purpose
        """

        url = "/api/v1/fileTokens"
        if purpose == 'UPLOAD':
            data = (
                {
                    "fileTokenType": "UPLOAD"
                }
            )
        elif purpose == 'DOWNLOAD':
            data = (
                {
                    "fileTokenType": "DOWNLOAD"
                }
            )
        else:
            return FortifyResponse(message='attribute purpose must be either UPLOAD or DOWNLOAD', success=False)

        return self._request('POST', url, json=data)

    def get_issue_template(self, project_template_id):
        """
        :param project_template_id: id of project template
        :return: A response object with data containing issue templates for the supplied project name
        """

        url = "/api/v1/issueTemplates" + "?limit=1&q=id:\"" + project_template_id + "\""
        return self._request('GET', url)

    def get_issue_template_id(self, project_template_name):
        """
        :param project_template_name: name of project template
        :return: A response object with data containing issue templates for the supplied project name
        """

        url = "/api/v1/issueTemplates" + "?limit=1&fields=id&q=name:\"" + project_template_name + "\""
        return self._request('GET', url)

    def get_project_version_artifacts(self, parent_id):
        """
        :param parent_id: parent resource identifier
        :return: A response object containing project version artifacts
        """
        url = "/api/v1/projectVersions/" + str(parent_id) + "/artifacts?start=0&limit=-1"
        return self._request('GET', url)

    def get_project_version_attributes(self, project_version_id):
        """
        :param project_version_id: Project version id
        :return: A response object containing the project version attributes
        """
        url = '/api/v1/projectVersions/' + str(project_version_id) + '/attributes/?start=0&limit=-1'
        return self._request('GET', url)

    def get_all_project_versions(self):
        """
        :return: A response object with data containing project versions
        """
        url = "/api/v1/projectVersions?start=0&limit=-1"
        return self._request('GET', url)

    def get_project_version(self, version_id):
        """
        :version_id: Project Version ID
        :return: Details of a Project Version
        """
        url = "/api/v1/projectVersions/" + version_id
        return self._request('GET', url)
    
    def get_project_versions(self, project_name):
        """
        Implemented on SSC as project-version-controller to manage application versions.
        A variety of associated resources are accessible via links.
        :return: A response object with data containing project versions
        """
        url = "/api/v1/projectVersions?limit=0&q=project.name:\"" + project_name + "\""
        return self._request('GET', url)

    def get_version(self, version_name):
        """
        :return: A response object with data containing just a project's version
        """
        url = "/api/v1/projectVersions?q=name:\"" + version_name + "\""
        return self._request('GET', url)
    
    def set_project_versions_test(self, project_name, project_version_name):
        """
        Check whether the specified application name is already defined in the system
        :param project_name: Application or Project name in SSC you want to test the value of.
        :param project_version_name: Application or Project version you want to test the value of.
        :return: A response object of found for true or false
        """
        data = {
            "projectName": project_name,
            "projectVersionName": project_version_name
        }

        url = "/api/v1/projectVersions/action/test"
        return self._request('POST', url, json=data)

    def set_projects_test(self, application_name):
        """
        Check whether the specified application name is already defined in the system.  For some reason SSC is not
        consistent here on naming conventions between project and application.
        :param application_name: Application or Project name in SSC you want to test the value of.
        :return: A response object of found for true or false
        """
        data = {
            "applicationName": application_name
        }

        url = "/api/v1/projects/action/test"
        return self._request('POST', url, json=data)

    def get_projects_id_from_name(self, project_name):
        """
        Don't want to trust user input here.  Encoding for spaces with + sign and other encoding
        :return: the data object from querying by a Project or Application name
        """

        url = "/api/v1/projects?q=name:" + urllib.parse.quote_plus(project_name)
        return self._request('GET', url)

    def get_projects(self):
        """
        :return: A response object with data containing projects
        """

        url = "/api/v1/projects?start=0&limit=-1"

        return self._request('GET', url)

    def get_token(self, description, type='UnifiedLoginToken'):
        """
        Token types include UnifiedLoginToken, UploadFileTransferToken, DownloadFileTransferToken
        :return: A response object with data containing create date, terminal date, and the actual token
        """
        data = {
            "description": description,
            "type": type
        }

        url = '/api/v1/tokens'
        return self._request('POST', url, json=data)

    def post_attribute_definition(self, attribute_definition):
        """
        :param attribute_definition:
        :return:
        """
        url = '/api/v1/attributeDefinitions'
        return self._request('POST', url, json=attribute_definition)

    def upload_artifact_scan(self, file_path, project_version_id):
        """
        :param file_path: full path to the file to upload
        :param project_version_id: project_version_id
        :return: Response from the file upload operation
        """
        upload = self.get_file_token('UPLOAD')
        if upload is None or upload.data['data'] is None:
            return FortifyResponse(message='Failed to get the SSC upload file token', success=False)

        file_token = upload.data['data']['token']
        url = "/upload/resultFileUpload.html?mat=" + file_token
        files = {'file': (ntpath.basename(file_path), open(file_path, 'rb'))}

        headers = {
            'Accept': 'Accept:application/xml, text/xml, */*; q=0.01',
            'Accept-Encoding': 'gzip, deflate, br',
            'Connection': 'keep-alive'
        }

        params = {
            'entityId': project_version_id,
            'clientVersion': self.client_version,
            'Upload': "Submit Query",
            'Filename': ntpath.basename(file_path)
        }

        return self._request('POST', url, params, files=files, stream=True, headers=headers)

    #TODO change to '/api/v1/coreRulepacks/', "file=@rule.xml;type=text/xml, 'Content-Type': 'multipart/form-data',
    def upload_rulepack(self, file_path):
        """
        Upload rulepack to Fortify SSC
        :param file_path:
        :return:
        """
        upload = self.get_file_token('UPLOAD')
        if upload is None or upload.data['data'] is None:
            return FortifyResponse(message='Failed to get the SSC upload file token', success=False)

        file_token = upload.data['data']['token']
        url = "/upload/rulepackUpload.html?mat=" + file_token
        files = {'file': (ntpath.basename(file_path), open(file_path, 'rb'))}

        headers = {
            'Accept': 'Accept:application/xml, text/xml, */*; q=0.01',
            'Accept-Encoding': 'gzip, deflate',
            'Connection': 'keep-alive'
        }

        params = {
            'clientVersion': self.client_version,
            'Filename': ntpath.basename(file_path)
        }

        return self._request('POST', url, params, files=files, stream=True, headers=headers)

    def delete_token(self, token_id):
        """
        Delete a token by ID from the auth-token-controller
        :param id:
        :return:
        """
        url = "/api/v1/tokens/" + str(token_id)
        return self._request('DELETE', url)

    def delete_all_user_tokens(self):
        """
        Delete all tokens by user from the auth-token-controller
        :return:
        """
        url = '/api/v1/tokens' + '?all=true'
        return self._request('DELETE', url)

    def get_all_tokens(self):
        """
        Get all tokens for all users
        :return:
        """
        url = "/api/v1/tokens?start=0&limit=-1"
        return self._request('GET', url)

    #TODO: fix expire_date to one year out
    def set_token(self, description, token_type, expire_date="2028-12-29T22:40:11.000+0000"):
        """
        Create any type of SSC token required
        :param description:
        :param token_type:
        :return:
        """
        data = {
            "description": description,
            "type": token_type,
            "terminalDate": expire_date
        }

        url = "/api/v1/tokens"
        return self._request('POST', url, json=data)

    def get_all_rulepacks(self):
        """
        List all rules on an SSC instance
        :return:
        """
        url = "/api/v1/coreRulepacks"
        return self._request('GET', url)

    def delete_rulepack(self, rulepack_id):
        """
        Delete a given rulepack by ID
        :param rulepack_id:
        :return:
        """
        url = "/api/v1/coreRulepacks/" + str(rulepack_id)
        return self._request('DELETE', url)

    def update_rulepacks(self):
        """
        Described as a Dolimport, update rulepacks from the public fortify server and return status with rulepacks
        updated.
        Update Fortify Stock Rulepacks
        :return:
        """
        url = "/api/v1/updateRulepacks"
        return self._request('GET', url)

    def get_all_issue_aging(self):
        """
         :return: Get total summary of applicationVersions, averageDaysToRemediate, averageDaysToReview, filesScanned,
         issuesPendingReview, issuesRemediated, linesOfCode, openIssues, openIssuesReviewed
         """
        url = "/api/v1/portlets/issueaging"
        return self._request('GET', url)

    def get_project_version_issues(self, version_id, orderby='friority'):
        """
        Issues per application/project version, with optional orderby query param
        :param version_id:
        :return: a List w/ no Project Version Name
        """
        url = '/api/v1/projectVersions/' + str(version_id) + '/issues?start=0&limit=-1&' + orderby + '&' \
                                                             'showhidden=false&' \
                                                             'showremoved=false&showsuppressed=false&' \
                                                             'showshortfilenames=false'
        return self._request('GET', url)

    def get_project_version_issue_details(self, issue_id):
        """
        Returns trace analysis and other details of a given issue.  The issue ID can be found from the /issues or
        projectVersions endpoint.
        :param issue_id:
        :return: full detail of a given issue
        """
        url = '/api/v1/issueDetails/' + str(issue_id)
        return self._request('GET', url)

    def get_project_version_source_file(self, version_id, path):
        """
        Returns an object with information on a source file from a specific project version.
        It includes the file content.
        :param version_id: application version id
        :param path: the relative path of the file from the project root folder
        :return: details on a source file
        """
        url = f'/api/v1/projectVersions/{version_id}/sourceFiles?q=path:"{path}"'
        return self._request("GET", url)

    def get_cloud_pool_list(self):
        """
        Get listing of all cloud pools
        :return:
        """
        url = '/api/v1/cloudpools'
        return self._request('GET', url)

    def get_cloud_worker_list(self):
        """
        Get listing of all cloud sensors/workers
        :return:
        """
        url = '/api/v1/cloudworkers'
        return self._request('GET', url)

    def set_cloud_pool(self, description, name):
        """
        Creates a cloudscan pool
        :param description:
        :param name:
        :return:
        """
        data = {
            "description": description,
            "name": name
        }
        url = '/api/v1/cloudpools'
        return self._request('POST', url, json=data)

    def set_cloud_worker(self, worker_uuid, pool_id):
        """
        Assignes a cloudscan worker to a pool
        :param pool_id: ID of cloudpool
        :param worker_uuid: uuid from the cloudworker
        :return:
        """
        data = {
                "workerUuids": [str(worker_uuid)]
        }
        url = '/cloudpools/' + pool_id + '/workers/action/assign'
        return self._request('POST', url, json=data)

    def get_all_ldap_users(self):
        """
        List all ldap users
        :return:
        """
        url = '/api/v1/ldapObjects'
        return self._request('GET', url)

    def set_ldap_user(self, distinguished_name, project_version_id):
        """
        Add LDAP user to Fortify SSC
        :return:
        """
        data = {"distinguishedName": distinguished_name,
                "roles": [{"permissionIds": [str(project_version_id)]}]
               }
        url = '/api/v1/ldapObjects'
        return self._request('POST', url, json=data)

    def get_all_auth_entities(self):
        """
        Manage aggregated list of authentication entities (local and LDAP user accounts). LDAP groups can be accessed
        via a linked resource. This controller does not support creation of new accounts. To create new local or LDAP
        user accounts, use the corresponding specific controllers.  In other words Give me a listing of all users local
        or LDAP, for identying the auth_entity_id to be used in get_auth_entity, and set_auth_entity_to_version
        :return:
        """
        url = '/api/v1/authEntities?fulltextsearch=false'
        return self._request('GET', url)

    def get_auth_entity(self, auth_entity_id):
        """
        Manage the association of application versions to the authentication entity (local or LDAP user)
        :param auth_entity_id:
        :return:
        """
        url = '/api/v1/authEntities/' + auth_entity_id + '/projectVersions'
        return self._request('GET', url)

    def modify_auth_entity_to_version(self, auth_entity_id, project_version_id):
        """
        TODO: NOT SURE IF NEEDED
        :param auth_entity_id:
        :param project_version_id:
        :return:
        """
        data = [project_version_id]
        url = '/api/v1/authEntities' + auth_entity_id + 'projectVersions'
        return self._request('PUT', url, json=data)

    def set_auth_entity_to_version(self, auth_entity_id, project_version_id):
        """
        Associate the specified application versions to the authentication entity
        :return:
        """
        data = {"projectVersionIds": [str(project_version_id)]}
        url = '/api/v1/authEntities' + auth_entity_id + 'projectVersions/action/assign'
        return self._request('POST', url, json=data)

    def get_all_auth_entity_of_project_version(self, project_version_id,
                                               extract_users_from_groups=True, include_universal_access_entities=True):
        """
        Retrieve the authentication entities associated with this application version.  What this really means is
        give me the users entitled for a given Project version
        :param project_version_id:
        :return: List all users associated with the Project Version
        """
        url = '/api/v1/projectVersions/' + str(project_version_id) + '/authEntities'
        params = {
            'extractusersfromgroups': extract_users_from_groups,
            'includeuniversalaccessentities': include_universal_access_entities
        }

        return self._request('GET', url, params)

    def _request(self, method, url, params=None, files=None, json=None, data=None, headers=None, stream=False):
        """Common handler for all HTTP requests."""
        if not params:
            params = {}

        if not headers:
            headers = {
                'Accept': 'application/json'
            }
            if method == 'GET' or method == 'POST' or method == 'PUT':
                headers.update({'Content-Type': 'application/json'})
        headers.update({'User-Agent': self.user_agent})

        try:

            if self.auth_type == 'basic':
                response = requests.request(method=method, url=self.host + url, params=params, files=files,
                                            headers=headers, json=json, data=data,
                                            timeout=self.timeout, verify=self.verify_ssl,
                                            auth=(self.username, self.password), stream=stream)
            elif self.auth_type == 'token':
                response = requests.request(method=method, url=self.host + url, params=params, files=files,
                                            headers=headers, json=json, data=data, timeout=self.timeout, 
                                            verify=self.verify_ssl, auth=FortifyTokenAuth(self.token), stream=stream)
            else:
                response = requests.request(method=method, url=self.host + url, params=params, files=files,
                                            headers=headers, json=json, data=data,
                                            timeout=self.timeout, verify=self.verify_ssl, stream=stream)

            try:
                response.raise_for_status()

                # two flavors of response are successful, GETs return 200, PUTs return 204 with empty response text
                response_code = response.status_code
                success = True if response_code // 100 == 2 else False
                if response.text:
                    try:
                        data = response.json()
                    except ValueError:  # Sometimes the returned data isn't JSON, so return raw
                        data = response.content

                return FortifyResponse(success=success, response_code=response_code, data=data,
                                       headers=response.headers)
            except ValueError as e:
                return FortifyResponse(success=False, message="JSON response could not be decoded {0}.".format(e))
        except requests.exceptions.SSLError as e:
            return FortifyResponse(message='An SSL error occurred. {0}'.format(e), success=False)
        except requests.exceptions.ConnectionError as e:
            return FortifyResponse(message='A connection error occurred. {0}'.format(e), success=False)
        except requests.exceptions.Timeout:
            return FortifyResponse(message='The request timed out after ' + str(self.timeout) + ' seconds.',
                                   success=False)
        except requests.exceptions.RequestException as e:
            return FortifyResponse(
                message='There was an error while handling the request. {0}'.format(e), success=False)


class FortifyTokenAuth(requests.auth.AuthBase):
    def __init__(self, token):
        self.token = token

    def __call__(self, r):
        r.headers['Authorization'] = 'FortifyToken ' + self.token
        return r


class FortifyResponse(object):
    """Container for all Fortify SSC API responses, even errors."""

    def __init__(self, success, message='OK', response_code=-1, data=None, headers=None):
        self.message = message
        self.success = success
        self.response_code = response_code
        self.data = data
        self.headers = headers

    def __str__(self):
        if self.data:
            return str(self.data)
        else:
            return self.message

    def data_json(self, pretty=False):
        """Returns the data as a valid JSON string."""
        if pretty:
            return json.dumps(self.data, sort_keys=True, indent=4, separators=(',', ': '))
        else:
            return json.dumps(self.data)



from typing import Union, Tuple
import time

from .exceptions import *
from .template import *
from .query import Query
from .api import FortifySSCAPI


class FortifySSCClient:

    def __init__(self, url: str, auth: Union[str, Tuple[str, str]], proxies=None, verify=True):
        """
        :param url: url to ssc, including the path. E.g. `https://fortifyssc/ssc`
        :param auth: Authentication, either a token str or a (username, password) tuple
        """
        self._url = url
        self._auth = auth
        self._api = FortifySSCAPI(url, auth, proxies, verify)

        self.projects = Project(self._api, None, self)
        self.pools = CloudPool(self._api, None, self)
        self.jobs = CloudJob(self._api, None, self)
        self.reports = Report(self._api, None, self)
        self.auth_entities = AuthEntity(self._api, None, self)
        self.ldap_user = LdapUser(self._api, None, self)

    def _list(self, endpoint, **kwargs):
        with self._api as api:
            for e in api.page_data(endpoint, **kwargs):
                yield e

    def list_engine_types(self, **kwargs):
        for e in self._list('/api/v1/engineTypes', **kwargs):
            yield Engine(self._api, e, self)

    def list_all_project_versions(self, **kwargs):
        kwargs['limit'] = -1
        for e in self._list('/api/v1/projectVersions', **kwargs):
            yield Version(self._api, e, None)

    @property
    def api(self):
        return self._api


class SSCObject(dict):
    def __init__(self, api, obj=None, parent=None):
        super().__init__(obj if obj else {})
        assert isinstance(api, FortifySSCAPI), 'Wrong parameter type, api should be FortifySSCAPI'
        self._api = api
        self.parent = parent

    def __str__(self):
        return f"{self.__class__}({super().__str__()})"

    def is_instance(self):
        return len(self) != 0

    def assert_is_instance(self, msg=None):
        if not self.is_instance():
            raise NotAnInstanceException(msg)


class Version(SSCObject):

    def __init__(self, api, obj, parent):
        super().__init__(api, obj, parent)
        self.attributes = Attribute(api, obj, self)
        self.issues = Issue(api, obj, self)
        self.custom_tags = CustomTag(api, obj, self)
        self.artifacts = Artifact(api, obj, self)

    def initialize(self, template=DefaultVersionTemplate):
        """
        Called automatically when Version.create is called.

        :return:
        """
        with self._api as api:
            if not isinstance(template, DefaultVersionTemplate):
                template = template()
            data = template.generate(api=api, project_version_id=self['id'])
            return api.bulk_request(data)

    def create(self, version_name, description="", active=True, committed=False, template=DefaultVersionTemplate):
        """ Creates a version for the CURRENT project """
        self.assert_is_instance("Cannot create version for empty project - consider using `create_project_version`")
        assert self.parent['name'] is not None, "how is the parent name None?"
        return self.parent.create(self.parent['name'], version_name, description=description, active=active,
                           committed=committed, project_id=self.parent['id'],
                           issue_template_id=self.parent['issueTemplateId'], template=template)

    def copy(self, new_name: str, new_description: str = ""):
        """
        Copy THIS project version including findings and finding state.
        Useful for some operations, e.g. pull requests
        """
        self.assert_is_instance()
        return self.create(new_name, new_description, active=self['active'], committed=self['committed'],
                           template=CloneVersionTemplate(self['id']))

    def list(self, **kwargs):
        if not self.parent:
            raise ParentNotFoundException("No project parent found to query versions from")
        with self._api as api:
            for e in api.page_data(f"/api/v1/projects/{self.parent['id']}/versions", **kwargs):
                p = Project(self._api, e['project'], None) if 'project' in e else self.parent
                yield Version(self._api, e, p)

    def list_auth_entities(self, **kwargs):
        self.assert_is_instance()
        with self._api as api:
            for e in api.page_data(f"/api/v1/projectVersions/{self['id']}/authEntities", **kwargs):
                yield AuthEntity(self._api, e, self.parent)

    def get(self, id):
        with self._api as api:
            return Version(self._api, api.get(f"/api/v1/projectVersions/{id}")['data'], self.parent)

    def delete(self):
        """ Delete the current version """
        self.assert_is_instance()
        with self._api as api:
            return api.delete(f"/api/v1/projectVersions/{self['id']}")

    def get_processing_rules(self, **kwargs):
        self.assert_is_instance()
        with self._api as api:
            return api.get(f"/api/v1/projectVersions/{self['id']}/resultProcessingRules", **kwargs)

    def set_processing_rules(self, rules):
        self.assert_is_instance()
        with self._api as api:
            return api.put(f"/api/v1/projectVersions/{self['id']}/resultProcessingRules", rules)

    def issue_summary(self, series_type='DEFAULT', group_axis_type='ISSUE_FOLDER'):
        self.assert_is_instance()
        with self._api as api:
            return api.get(f"/api/v1/projectVersions/{self['id']}/issueSummaries",
                           seriestype=series_type,
                           groupaxistype=group_axis_type)['data']


    def upload_artifact(self, file_path, process_block=False):
        """
        :param process_block: Block this method for Artifact processing
        """
        self.assert_is_instance()
        with self._api as api:
            with open(file_path, 'rb') as f:
                robj = api._request('POST', f"/api/v1/projectVersions/{self['id']}/artifacts", files={'file': f})
                art = Artifact(self._api, robj['data'], self)
                if process_block:
                    while a := art.get(art['id']):
                        if a['status'] in ['PROCESS_COMPLETE', 'ERROR_PROCESSING', 'REQUIRE_AUTH']:
                            return a
                        time.sleep(1)
                return art


class Project(SSCObject):

    def __init__(self, api, obj=None, parent=None):
        super().__init__(api, obj, parent)
        self.versions = Version(api, None, self)

    def list(self, **kwargs):
        """
        :param kwargs: The request query parameters
        :returns: Generator of each fortifyapi.Project
        """
        with self._api as api:
            for e in api.page_data('/api/v1/projects', **kwargs):
                yield Project(self._api, e, self.parent)

    def test(self, application_name: str) -> bool:
        """
        Check whether the specified application name is already defined in the system
        :returns: If the application_name was found
        """
        with self._api as api:
            return api.post(f"/api/v1/projects/action/test", applicationName=application_name)['data']['found']

    def get(self, id):
        """
        Get the given project by id
        :param id:
        :returns: fortifyapi.Project
        """
        with self._api as api:
            return Project(self._api, api.get(f"/api/v1/projects/{id}")['data'], self.parent)

    def update(self):
        """
        :returns: The updated Project object
        """
        self.assert_is_instance()
        with self._api as api:
            return Project(self._api, api.put(f"/api/v1/projects/{self['id']}", self)['data'], self)

    def create(self, project_name, version_name, project_id=None, description="", active=True,
               committed=False, issue_template_id='Prioritized-HighRisk-Project-Template',
               template=DefaultVersionTemplate):
        """

        :param project_name:
        :param version_name:
        :param project_id:
        :param description:
        :param active:
        :param committed:
        :param issue_template_id:
        :param template:
        :return: Returns the Version object
        :rtype: fortifyapi.Version
        """
        with self._api as api:
            r = api.post(f"/api/v1/projectVersions", {
                'name': version_name,
                'description': description,
                'active': active,
                'committed': committed,
                'project': {
                    'id': project_id,  # if this is None it will create the project
                    'name': project_name,
                    'description': description,
                    'issueTemplateId': issue_template_id
                },
                'issueTemplateId': issue_template_id
            })
            p = Project(self._api, r['data']['project'], None) if 'project' in r['data'] else self

            v = Version(self._api, r['data'], p)
            v.initialize(template=template)
            # get it again so we see it's true state
            # but we should really just re-get the Project so it has all the proper data
            p = Project(self._api, {}, None).get(p['id'])
            return p.versions.get(v['id'])

    def upsert(self, project_name, version_name, description="", active=True,
               committed=False, issue_template_id='Prioritized-HighRisk-Project-Template',
               template=DefaultVersionTemplate) -> Version:
        """ same as create but uses existing project and version"""
        # see if the project exists
        # TODO: change this to /projectVersions/action/test with {projectName:x, projectVersionName: y}
        q = Query().query("name", project_name)
        projects = list(self.list(q=q))
        if len(projects) == 0:
            return self.create(project_name, version_name, description=description, active=active, committed=committed,
                               issue_template_id=issue_template_id, template=template)
        else:
            # should be the first one
            project = projects[0]
            # but check if the version is there...
            for v in project.versions.list():
                if v['name'] == version_name:
                    return v
            return self.create(project_name, version_name, project_id=project['id'], description=description, active=active,
                               committed=committed, issue_template_id=issue_template_id, template=template)

    def delete(self):
        # delete every version and project will delete
        self.assert_is_instance()
        for v in self.versions.list():
            v.delete()


class Engine(SSCObject):
    pass


class CloudPool(SSCObject):

    def get(self, uuid):
        with self._api as api:
            return CloudPool(self._api, api.get(f"/api/v1/cloudpools/{uuid}")['data'], self.parent)

    def list(self, **kwargs):
        with self._api as api:
            for e in api.page_data(f"/api/v1/cloudpools", **kwargs):
                yield CloudPool(self._api, e, self.parent)

    def create(self, pool_name):
        with self._api as api:
            r = api.post(f"/api/v1/cloudpools", {
                "name": pool_name
            })
            return CloudPool(self._api, r['data'])

    def delete(self):
        self.assert_is_instance()
        with self._api as api:
            return api.delete(f"/api/v1/cloudpools/{self['uuid']}")

    def assign(self, worker_uuids):
        self.assert_is_instance()
        if not isinstance(worker_uuids, list):
            worker_uuids = [worker_uuids]
        with self._api as api:
            r = api.post(f"/api/v1/cloudpools/{self['uuid']}/workers/action/assign", {
                "workerUuids": worker_uuids
            })
            #TODO: figure out what this returns

    def jobs(self):
        self.assert_is_instance()
        with self._api as api:
            for e in api.page_data(f"/api/v1/cloudpools/{self['uuid']}/jobs"):
                yield CloudJob(self._api, e, self)


class CloudWorker(SSCObject):

    def list_unassigned(self, **kwargs):
        with self._api as api:
            for e in api.page_data(f"/api/v1/cloudpools/disabledWorkers", **kwargs):
                yield CloudWorker(self._api, e, self.parent)


class CloudJob(SSCObject):

    def list(self, **kwargs):
        with self._api as api:
            for e in api.page_data(f"/api/v1/cloudjobs", **kwargs):
                yield CloudJob(self._api, e, self.parent)

    def list_all(self, **kwargs):
        """ Helper function to just disable paging and get them all """
        kwargs['limit'] = -1
        for e in self.list(**kwargs):
            yield e

    def get(self, job_token):
        with self._api as api:
            return CloudJob(self._api, api.get(f"/api/v1/cloudjobs/{job_token}")['data'], self.parent)

    def cancel(self):
        with self._api as api:
            return api.post(f"/api/v1/cloudjobs/action/cancel", jobTokens=[self['jobToken']])


class Scan(SSCObject):

    def get(self, id):
        f"/api/v1/scans/{id}" # GET

    def list(self):
        with self._api as api:
            for e in api.page_data(f"/api/v1/artifacts/{self['id']}/scans", **kwargs):
                yield Scan(self._api, e, self)


class Artifact(SSCObject):

    def __init__(self, api, obj=None, parent=None):
        super().__init__(api, obj, parent)
        self.scans = Scan(api, None, self)

    def get(self, id):
        with self._api as api:
            return Artifact(self._api, api.get(f"/api/v1/artifacts/{id}")['data'], self.parent)

    def list(self, **kwargs):
        self.assert_is_instance()
        with self._api as api:
            for e in api.page_data(f"/api/v1/projectVersions/{self['id']}/artifacts", **kwargs):
                yield Artifact(self._api, e, self)

    def delete(self, id):
        f"/api/v1/artifacts/{id}" # DELETE
        raise NotImplementedError()

    def approve(self):
        f"/api/v1/artifacts/action/approve" # POST
        raise NotImplementedError()

    def purge(self):
        f"/api/v1/artifacts/action/purge" # POST
        raise NotImplementedError()


class Issue(SSCObject):
    NOT_AN_ISSUE = 0
    RELIABILITY_ISSUE = 1
    BAD_PRACTICE = 2
    SUSPICIOUS = 3
    EXPLOITABLE = 4

    def list(self, **kwargs):
        with self._api as api:
            for e in api.page_data(f"/api/v1/projectVersions/{self.parent['id']}/issues", **kwargs):
                yield Issue(self._api, e, self.parent)

    def get(self, id):
        with self._api as api:
            return Issue(self._api, api.get(f"/api/v1/projectVersions/{self.parent['id']}/issues/{id}")['data'], self.parent)

    def assign(self, user):
        """
        :param user: user id?
        """
        self.assert_is_instance()
        o = dict(user=user)
        f"/api/v1/projectVersions/{self.parent['id']}/issues/action/assignUser"

    def audit(self,  analysis, comment="via automation", user=None, suppressed=False, tags=None):
        """
        :param analysis: zero to four
        :param comment: Issue comment
        :param user: Username to assign, else None
        :param suppressed: Will suppress the issue
        :param tags: ?
        """
        self.assert_is_instance()
        assert analysis in [
            Issue.NOT_AN_ISSUE,
            Issue.RELIABILITY_ISSUE,
            Issue.BAD_PRACTICE,
            Issue.SUSPICIOUS,
            Issue.EXPLOITABLE
        ], "Not a valid analysis type"
        o = {
            'issues': [{
                'id': self['id'],
                'revision': self['revision']
            }],
            'comment': comment,
            'suppressed': suppressed,
            'customTagAudit': [{
                'customTagGuid': '87f2364f-dcd4-49e6-861d-f8d3f351686b',
                'newCustomTagIndex': analysis
            }]
        }
        if user:
            o['user'] = user
        if tags:
            if isinstance(tags, list):
                for t in tags:
                    o['customTagAudit'].append(t)
            else:
                o['customTagAudit'].append(tags)

        with self._api as api:
            return api.post(f"/api/v1/projectVersions/{self.parent['id']}/issues/action/audit", o)

    def suppress(self, suppressed=True):
        self.assert_is_instance()
        o = {
            'issues': [{
                'id': self['id'],
                'revision': self['revision']
            }],
            "suppressed": suppressed
        }
        with self._api as api:
            return api.post(f"/api/v1/projectVersions/{self.parent['id']}/issues/action/suppress", o)

    def unsuppress(self):
        self.suppress(False)


class Attachment(SSCObject):

    def list(self, **kwargs):
        with self._api as api:
            for e in api.page_data(f"/api/v1/issues/{self.id}/attachments", **kwargs):
                yield Attachment(self._api, e)

    def get(self, id):
        f"/api/v1/projectVersions/{self.id}/attributes/{id}" # GET
        raise NotImplementedError()

    def update(self):
        f"/api/v1/projectVersions/{self.parent.id}/attributes/{self.id}" # UPDATE
        raise NotImplementedError()

    def delete(self, id):
        f"/api/v1/projectVersions/{self.parent.id}/attributes/{self.id}" # UPDATE
        raise NotImplementedError()

    def upload(self):
        f"/api/v1/issues/{self.id}/attachments" # POST
        raise NotImplementedError()

    def delete_all(self):
        f"/api/v1/issues/{self.parent.id}/attachments" # DELETE
        raise NotImplementedError()


class Attribute(SSCObject):

    def list(self, **kwargs):
        with self._api as api:
            for e in api.page_data(f"/api/v1/projectVersions/{self['id']}/attributes", **kwargs):
                yield Attribute(self._api, e, self.parent)

    def get(self, id):
        f"/api/v1/projectVersions/{self['id']}/attributes/{id}" # GET
        raise NotImplementedError()

    def create(self):
        f"/api/v1/projectVersions/{self['id']}/attributes" # POST
        raise NotImplementedError()

    def update(self):
        f"/api/v1/projectVersions/{self['id']}/attributes" # PUT
        raise NotImplementedError()


class Report(SSCObject):

    def list(self, **kwargs):
        with self._api as api:
            for e in api.page_data(f"/api/v1/reports", **kwargs):
                yield Report(self._api, e, self.parent)

    def get(self, id):
        f"/api/v1/reports/{id}"
        raise NotImplementedError()

    def create(self):
        f"/api/v1/reports" # POST
        raise NotImplementedError()

    def delete(self):
        """ Delete the current report """
        with self._api as api:
            return api.delete(f"/api/v1/reports/{self['id']}")


class FileToken(SSCObject):

    def create(self, purpose='DOWNLOAD'):
        """
        :param purpose: valid values are `DOWNLOAD` and `UPLOAD`
        """
        assert purpose in ['UPLOAD', 'DOWNLOAD'], "Unsupported purpose"

        with self._api as api:
            return FileToken(self._api, api.post(f"/api/v1/fileTokens"), self.parent)

    def delete(self):
        assert False, "Have not tested this"
        with self._api as api:
            return api.delete(f"/api/v1/fileTokens", self['id'])


class Token(SSCObject):

    def list(self, **kwargs):
        f"/api/v1/tokens" # GET
        raise NotImplementedError()

    def create(self, **kwargs):
        f"/api/v1/tokens"  # POST
        raise NotImplementedError()

    def update(self):
        f"/api/v1/tokens/{self['id']}"  # PUT
        raise NotImplementedError()

    def delete(self):
        f"/api/v1/tokens/{self['id']}"  # DELETE
        raise NotImplementedError()

    def revoke(self):
        f"/api/v1/tokens/revoke"  # POST with body
        raise NotImplementedError()


class Rulepack(SSCObject):

    def list(self, **kwargs):
        with self._api as api:
            for e in api.page_data(f"/api/v1/coreRulepacks", **kwargs):
                yield Rulepack(self._api, e, self.parent)

    def upload(self):
        f"/api/v1/coreRulepacks" # POST
        raise NotImplementedError()

    def delete(self):
        f"/api/v1/coreRulepacks/{self['id']}"  # DELETE
        raise NotImplementedError()


class CustomTag(SSCObject):
    """ Specifically for project versions """

    def list(self, **kwargs):
        with self._api as api:
            for e in api.page_data(f"/api/v1/projectVersions/{self.parent['id']}/customTags", **kwargs):
                yield CustomTag(self._api, e, self.parent)

    def create(self, **kwargs):
        f"/api/v1/projectVersions/{self.parent['id']}/customTags"
        raise NotImplementedError()

    def update(self):
        f"/api/v1/projectVersions/{self.parent['id']}/customTags/{self['id']}"
        raise NotImplementedError()

class Role(SSCObject):

    def list(self, **kwargs):
        with self._api as api:
            for e in api.page_data(f"/api/v1/roles", **kwargs):
                yield Roles(self._api, e, self.parent)


class AuthEntity(SSCObject):

    def list(self, **kwargs):
        with self._api as api:
            for e in api.page_data(f"/api/v1/authEntities", **kwargs):
                yield AuthEntity(self._api, e, self.parent)

    def get(self, id):
        with self._api as api:
            return api.get(f"/api/v1/authEntities/{id}", **kwargs)['data']

    def find_ldap_user(self, username):
        with self._api as api:
            data = api.get(f"/api/v1/authEntities", q="isLdap:true",
                                                embed='roles(name)', entityName=username, orderby='entityName',
                                                start=0, limit=-1)['data']
            if len(data) > 0:
                return AuthEntity(self._api, data[0], self.parent)
            return None

    def assign_to_versions(self, versions):
        """
        :rtype boolean: Succeess
        """
        self.assert_is_instance()
        if type(versions) is list or type(versions) is tuple:
            cva = versions
        else:
            cva = [versions]
        with self._api as api:
            return api.post(f"/api/v1/authEntities/{self['id']}/projectVersions/action", {
                "type": "assign",
                "ids": cva
            })['data']['status'] == 'success'


class LocalGroup(SSCObject):
    pass


class User(SSCObject):

    def get(self, username):
        pass


class LocalUser(SSCObject):
    pass


class LdapUser(SSCObject):

    def list(self, ldaptype='USER', **kwargs):
        kwargs['ldaptype'] = ldaptype
        with self._api as api:
            for e in api.page_data(f"/api/v1/ldapObjects", **kwargs):
                yield LdapUser(self._api, e, self.parent)

    def add(self, roles=None):
        self.assert_is_instance()
        if roles:
            self['roles'] = roles
        elif 'roles' not in self:
            self['roles'] = [{'id': 'developer'}]
        with self._api as api:
            return LdapUser(self._api, api.post(f"/api/v1/ldapObjects", self)['data'], self)
